/* An implementation of the heap data structure for Python.

   Danny Yoo (dyoo@hkn.eecs.berkeley.edu)

   Heaps are a very useful data structure, and convenient for representing
   priority queues.  They're also the heart of a sort routine called HeapSort
   that guarantees, in the worst case, O(n log(n)) performance.  Not bad for
   an overgrown array.  *grin*

   This is meant to be a small case study in writing an Python extension
   module, so I'll try to comment this code as thoroughly as I can.  I'll try
   to follow the design of a new object type, like the "noddy" type that the
   "Extending and Embedding" document talks about:

        http://python.org/doc/current/ext/dnt-basics.html

   There are a lot of good resources on the theory and implementation of
   heaps.  The one that I'm using to implement this structure is "Chapter 14:
   Heaps" of Jon Bentley's "Programming Pearls".

   By the way, functions that have the substring '_int_' in their names are
   meant for internal use only, so you won't see them mapped directly as
   methods.

**/


#include "Python.h"


/* Let's first define what we need to keep in a PyHeapObject. */
typedef struct {
	PyObject_HEAD
	PyObject** elements;
	PyObject* cmp_function;
	long size;
	long capacity;
} PyHeap_PyHeapObject;



/* Every new Python object needs to have a common structure at the very
   beginning of the structure definition.  The PyObject_HEAD part allows all
   functions in Python to handle PyObjects uniformly, and is a common way for
   C programmers to implement a kind of inheritence.

   See Include/object.h for more information about PyObject_HEAD. */


/* Python has a special macro for doing forward declarations of structures.
   We'll need to use it if we want to put the definition of PyHeap_PyHeapType
   after the PyHeapObject constructor. */
staticforward PyTypeObject PyHeap_PyHeapType;


/* We do some comparisons in PyHeap_int_isGreater() to see if a comparison is
   positive.  We precache zero in the static variable PyZero so we can do the
   numeric comparison.  PyZero is initialized in initPyHeap(), and will be a
   global for our purposes.  It's a hack.  *grin*   */
static PyObject* PyHeap_PyZero;



/* Let's explain how we want to construct a heap for priority queue
   situations. */
static PyObject* PyHeap_new(PyObject *module, PyObject *args,
			    PyObject *kwargs) {
	PyObject *cmp_function = NULL;
	PyObject *bltinmod = NULL;
	PyHeap_PyHeapObject *self = NULL;

	/* We want to take in an optional "comparision function" when we
	   construct this heap.  See:

	   http://python.org/doc/current/ext/parseTuple.html

	   for more details on PyArg_ParseTuple().  We use the variation that
	   supports keywords, but we could have just as easily used
	   PyArg_ParseTuple() instead.
	*/
	static char *kwlist[] = {"cmp", NULL};
	if (PyArg_ParseTupleAndKeywords(args, kwargs, "|O:PyHeap", kwlist,
					&cmp_function)) {
		/* If we have a comparison function, let's make sure we can
		   call it.  If problems occur, we'll just abandon
		   construction and raise an error. */
		if (cmp_function && !PyCallable_Check(cmp_function)) {
			PyErr_SetString(PyExc_TypeError,
					"parameter must be callable");
			return NULL;
		}
	}
	/* If no cmp had been given to us, let's just use the standard
           __builtin__.cmp() function. */
	if(cmp_function == NULL) {
		bltinmod = PyImport_ImportModule("__builtin__");
		cmp_function = PyDict_GetItemString(PyModule_GetDict(bltinmod),
						    "cmp");
		Py_DECREF(bltinmod);
	}
	/* Now that we know what our function is, we can build our
           instance. */
	self = PyObject_New(PyHeap_PyHeapObject, &PyHeap_PyHeapType);
	self->cmp_function = cmp_function;
	Py_INCREF(self->cmp_function);
	/* The INCREF here is important, since we want to keep our own
           ownership of the function.  Don't forget: Python functions are
           objects too! */
	self->size = 0;
	self->capacity = 1;
	self->elements = PyMem_New(PyObject*, 1);
	/* The heap structure, for efficiency reasons, asks us to sacrifice
           the first element as a "sentinel".  No problem; let's just set it
           to zero. */
	self->elements[0] = Py_BuildValue("i", 0);
	return (PyObject*)self;
}


/* Let's also explain how one would destroy one of these heaps.

    Notice that we Py_DECREF all of our stored object references, so that we
    don't leave anything hanging. */
static void PyHeap_delete(PyHeap_PyHeapObject* self) {
	long i;
	/* The range of indices here isn't a mistake.  I really do mean to use
           "<=" here because heaps are arrays that are indexed starting at 1.
           */
	for(i = 0; i <= self->size; i++)
		Py_DECREF(self->elements[i]);
	PyMem_Del(self->elements);
	Py_DECREF(self->cmp_function);
	PyObject_DEL(self);
}




/* We often need to expand the number of elements in our heap to avoid growing
   beyond capacity.

   Note: this is just for internal use; I don't expect to expose this
   function to normal users.

   If bad things (aka out-of-memory situations) happen, return -1.  Hopefully,
   our caller will be responsible enough to call PyErr_NoMemory() for us.
 */
static int PyHeap_int_expand(PyHeap_PyHeapObject *heap) {
	long doubled_capacity = heap->capacity * 2;
	PyObject** new_elements = PyMem_Resize(heap->elements,
					       PyObject*,
					       doubled_capacity);
	/** Check to make sure that new_elements isn't NULL.  If it is,
            we're in serious trouble, and should raise PyMem_NoMemory!  **/
	if (new_elements == NULL)
		return -1;
	heap->elements = new_elements;
	heap->capacity = doubled_capacity;
	return 0;
}



/* Here is a small convenience function to drive the comparison function.
   It returns 1 if a > b, 0 otherwise.

   We use a few functions from the Object protocols:
       http://python.org/doc/current/api/object.html
   as well as the Numeric protocols:
       http://python.org/doc/current/api/number.html */

static int PyHeap_int_isGreater(PyObject *cmp_func, PyObject *a, PyObject *b) {
	PyObject *args = NULL;
	PyObject *cmpresult = NULL;
	int cmpresult2 = 0;
	if ((args = Py_BuildValue("OO", a, b)) == NULL)
		goto error;
	if ((cmpresult = PyObject_CallObject(cmp_func, args)) == NULL)
		goto error;  
	cmpresult2 = PyObject_Compare(cmpresult, PyHeap_PyZero) > 0;
	Py_DECREF(args);
	Py_DECREF(cmpresult);
	return cmpresult2;
	/* At any point, if things mess up, let's clean up our reference
	   counts and return 0.  Py_XDECREF is a "safe" version of Py_DECREF
	   that doesn't die if args or cmpresult are still NULL, and are often
	   preferred when doing error cleanup. */
  error:
	Py_XDECREF(args);
	Py_XDECREF(cmpresult);
	return 0;
}




/***********************************************************************/

/* For the moment, let's work on the algorithm stuff that's specific to heaps.
   We explain how to "siftup" and "siftdown" newly added elements to a heap.

   This is basically a translation of Jon Bentley's C++ code, with a little
   mix of Knuth's optimization during the swapping of parent and child. 

   [fixme -- add a summary of the heap structure here]
*/

static void PyHeap_int_siftup(PyHeap_PyHeapObject* self, long n) {
	PyObject *obj = self->elements[n];
	while (n > 1 
	       && PyHeap_int_isGreater(self->cmp_function, 
				   obj, self->elements[n / 2])) {
		self->elements[n] = self->elements[n / 2];
		n = n / 2;
	}
	self->elements[n] = obj;
}


static void PyHeap_int_siftdown(PyHeap_PyHeapObject* self, long n) {
	PyObject *obj = self->elements[1];
	long i, c;
	for (i = 1; (c = 2*i) <= n; i = c) {
		if (c+1 <= n && PyHeap_int_isGreater(self->cmp_function,
						 self->elements[c+1], 
						 self->elements[c]))
			c++;
		if (PyHeap_int_isGreater(self->cmp_function,
				     obj,
				     self->elements[c]))
			break;
		self->elements[i] = self->elements[c];
	}
	self->elements[i] = obj;
}


static int PyHeap_int_insert(PyHeap_PyHeapObject* self, PyObject *obj) {
	if (self->size + 1 == self->capacity) {
		if(PyHeap_int_expand(self) == -1) { return -1; }
	}
	self->size++;
	self->elements[self->size] = obj;
	Py_INCREF(obj);
	PyHeap_int_siftup(self, self->size);
	return 0;
}

/***********************************************************************/

/* Back to Python related stuff.  All of these definitions are useless if we
   don't allow the user to get at them.  Let's define some of the functions
   our users can touch. */


static PyObject* PyHeap_getSize(PyHeap_PyHeapObject* heap, PyObject* args) {
	if(PyArg_ParseTuple(args,"")) {
		return Py_BuildValue("i", heap->size);
	}
	/* Otherwise, the PyArg_ParseTuple() has already set the exception
	   stuff up, so we can return NULL with no problems. */
	return NULL;
}

static PyObject* PyHeap_push(PyHeap_PyHeapObject* heap, PyObject *args) {
	PyObject *obj = NULL;
	if(PyArg_ParseTuple(args, "O", &obj)) {
		if(PyHeap_int_insert(heap, obj) == -1) { 
			return PyErr_NoMemory();
		}
		/* Even a function that doesn't "return" anything in Python
		   still returns Py_None. */
		Py_INCREF(Py_None);
		return Py_None;
	}
	return NULL;
}

static PyObject* PyHeap_pop(PyHeap_PyHeapObject* heap, PyObject *args) {
	PyObject *obj = NULL;
	if (PyArg_ParseTuple(args, "")) {
		if (heap->size > 0) {
		obj = heap->elements[1];
		heap->elements[1] = heap->elements[heap->size--];
		PyHeap_int_siftdown(heap, heap->size);
		return obj;
		}
		else {
			/* If the heap is empty, raise an error here. */
			PyErr_SetString(PyExc_IndexError,
					"pop from empty heap");
			return NULL;
		}
	}
	return NULL;
}



/* Now that we have all these methods, we need to communicate these methods to
   Python.  Notice that all the functions above were static... so their names
   are scrambled at compile time to avoid naming conflicts when we link up
   with another C program.  This is usually a good thing, except we really do
   want Python to be able to touch our functions!

   This dilemma is solved if we pass Python a function table of the methods we
   want to expose to it.  Here we tell Python what the methods of our
   PyHeapObject. */
static PyMethodDef PyHeapObjectMethods[] = {

	{"size", (PyCFunction)PyHeap_getSize, METH_VARARGS,
	 "Return the number of elements in the heap."},
	{"push", (PyCFunction)PyHeap_push, METH_VARARGS,
	 "Push an element onto the heap." },
	{"pop", (PyCFunction)PyHeap_pop, METH_VARARGS,
	 "Pop the highest-priority item off the heap."},
	{NULL, NULL, 0, NULL}    /* Sentinel */
};

static PyObject* PyHeap_getattr(PyHeap_PyHeapObject* self, char* name) {
	return Py_FindMethod(PyHeapObjectMethods, (PyObject*)self, name);
}

static PyTypeObject PyHeap_PyHeapType = {
	PyObject_HEAD_INIT(NULL)
	0,
	"PyHeap",
	sizeof(PyHeap_PyHeapObject),
	0,
	(destructor) PyHeap_delete,  /*tp_dealloc*/
	0,              /*tp_print*/
	(getattrfunc) PyHeap_getattr, /*tp_getattr*/
	0,              /*tp_setattr*/
	0,              /*tp_compare*/
	0,              /*tp_repr*/
	0,              /*tp_as_number*/
	0,              /*tp_as_sequence*/
	0,              /*tp_as_mapping*/
	0,              /*tp_hash */
};



/* Whew!  We also need to add our PyHeapObject constructor as a module-level
   function.  That's done here: */
static PyMethodDef PyHeapModuleMethods[] = {
	{"PyHeap", (PyCFunction)PyHeap_new, METH_VARARGS | METH_KEYWORDS,
	 "Create a new PyHeap object."},
	{NULL, NULL, 0, NULL}
};


/* Finally, here's our module initialization routine. */
DL_EXPORT(void)
	initPyHeap(void) {
	PyHeap_PyHeapType.ob_type = &PyType_Type;
	Py_InitModule("PyHeap", PyHeapModuleMethods);
	PyHeap_PyZero = Py_BuildValue("i", 0);
}
